import numpy as np
import xarray as xr
from numba import jit


def lateral_fill(
    da_in, isvalid_mask, ltripole=False, tol=1.0e-4, use_sor=False, rc=1.8, max_iter=10000
):
    """Perform lateral fill on xarray.DataArray

    Parameters
    ----------
    da_in : xarray.DataArray
      DataArray on which to fill NaNs. Fill is performed on the two
      rightmost dimenions. Grid is assumed periodic in `x` direction
      (last dimension).

    isvalid_mask : xarray.DataArray, boolean
      Valid values mask: `True` where data should be filled. Must have the
      same rightmost dimenions as `da_in`.

    ltripole : boolean, optional [default=False]
      Logical flag; if `True` then treat the top row of the grid as periodic
      in the sense of a tripole grid.

     tol : float, optional [default=1.0e-4]
      Convergence criteria: stop filling when values change is less or equal
      to `tol * var`; i.e. `delta <= tol * np.abs(var[j, i])`.

    use_sor: boolean, optional [default=False]
      switch to select SOR fill algorithm over progressive fill algorithm

    rc : float, optional [default=1.8, valid bounds=(1.0,2.0)]
       over-relaxation coefficient to use in SOR fill algorithm.
       Larger arrrays (or extent of region to be filled if not global)
       typically converge faster with larger coefficients.
       For completely land-filling a 1 deg. grid (360x180) a coefficient in
       the range 1.85-1.9 is near optimal.

    max_iter : integer, optional, [default=10000]
       maximum number of iterations to do before giving up if tol is not reached.

    Returns
    -------
    da_out : xarray.DataArray
      DataArray with NaNs filled by iterative smoothing.

    """
    dims_in = da_in.dims
    non_lateral_dims = dims_in[:-2]

    attrs = da_in.attrs
    encoding = da_in.encoding
    coords = da_in.coords

    da_in, isvalid_mask = xr.broadcast(da_in, isvalid_mask)

    if len(non_lateral_dims) > 0:
        da_in_stack = da_in.stack(non_lateral_dims=non_lateral_dims)
        da_out_stack = xr.full_like(da_in_stack, fill_value=np.nan)
        isvalid_mask_stack = isvalid_mask.stack(non_lateral_dims=non_lateral_dims)
        for i in range(da_in_stack.shape[-1]):
            arr = da_in_stack.data[:, :, i]
            da_out_stack[:, :, i] = lateral_fill_np_array(
                arr, isvalid_mask_stack.data[:, :, i], ltripole, tol, use_sor, rc, max_iter
            )

        da_out = da_out_stack.unstack('non_lateral_dims').transpose(*dims_in)

    else:
        da_out = xr.full_like(da_in, fill_value=np.nan)
        da_out[:, :] = lateral_fill_np_array(
            da_in.data, isvalid_mask.data, ltripole, tol, use_sor, rc, max_iter
        )

    da_out.attrs = attrs
    da_out.encoding = encoding
    for k, da in coords.items():
        da_out[k].attrs = da.attrs

    return da_out


def lateral_fill_np_array(
    var, isvalid_mask, ltripole=False, tol=1.0e-4, use_sor=False, rc=1.8, max_iter=10000
):

    """Perform lateral fill on numpy.array

    Parameters [NB defaults set redundantly with lateral_fill above to allow this
                function to be called directly for numpy arrays]
    ----------

    var : numpy.array
      Array on which to fill NaNs. Fill is performed on the two
      rightmost dimenions. Grid is assumed periodic in `x` direction
      (last dimension). Only NaNs where isvalid_mask is True will be filled.

    isvalid_mask : numpy.array, boolean
      Valid values mask: `True` where data should be filled. Must have the
      same rightmost dimenions as `da_in`.

    ltripole : boolean, optional [default=False set in lateral_fill]
      Logical flag; if `True` then treat the top row of the grid as periodic
      in the sense of a tripole grid.

    tol : float, optional [default=1.0e-4 set in lateral_fill]
      Convergence criteria: stop filling when values change is less or equal
      to `tol * var`; i.e. `delta <= tol * np.abs(var[j, i])`.

    use_sor: boolean, optional [default=False set in lateral_fill]
      switch to select SOR fill algorithm

    rc : float, optional [default=1.8, valid bounds=(1.0,2.0)]
       over-relaxation coefficient to use in SOR fill algorithm.
       Larger arrrays (or extent of region to be filled if not global)
       typically converge faster with larger coefficients.
       For completely land-filling a 1 deg. grid (360x180) a coefficient in
       the range 1.85-1.9 is near optimal.

    max_iter : integer, optional, [default=10000 set in lateral_fill]
       maximum number of iterations to do before giving up if tol is not reached.

    Returns
    -------

    da_out : xarray.DataArray
      DataArray with NaNs filled by iterative smoothing.

    """

    assert var.ndim == 2, f'lateral_fill_np_array expects 2D array, but got {var.ndim}: {var.shape}'

    nlat, nlon = var.shape[-2:]

    var = var.copy()

    if use_sor:
        fillmask = np.isnan(var)  # Fill all NaNs
        keepNaNs = ~isvalid_mask & np.isnan(var)
        _iterative_fill_sor(nlat, nlon, var, fillmask, tol, rc, max_iter, ltripole)
        var[keepNaNs] = np.nan  # Replace NaNs in areas not designated for filling
    else:
        fillmask = np.isnan(var) & isvalid_mask
        _iterative_fill_POP_core(nlat, nlon, var, fillmask, tol, ltripole, max_iter)

    return var


@jit(nopython=True, parallel=True)
def _iterative_fill_POP_core(nlat, nlon, var, fillmask, tol, ltripole, max_iter):
    """Iterative smoothing algorithm."""

    done = False
    iter_cnt = 0

    work = np.empty_like(var)

    while not done:
        done = True
        iter_cnt += 1

        # assume bottom row is land, so skip it
        for j in range(1, nlat):
            jm1 = j - 1
            jp1 = j + 1

            for i in range(0, nlon):
                # assume periodic in x
                im1 = (i - 1) % nlon
                ip1 = (i + 1) % nlon

                work[j, i] = var[j, i]

                if not fillmask[j, i]:
                    continue

                numer = 0.0
                denom = 0.0

                # East
                if np.isfinite(var[j, ip1]):
                    numer += var[j, ip1]
                    denom += 1.0

                # North
                if j < nlat - 1:
                    if np.isfinite(var[jp1, i]):
                        numer += var[jp1, i]
                        denom += 1.0

                else:
                    # assume only tripole has non-land top row
                    if ltripole:
                        if np.isfinite(var[j, nlon - 1 - i]):
                            numer += var[j, nlon - 1 - i]
                            denom += 1.0

                # West
                if np.isfinite(var[j, im1]):
                    numer += var[j, im1]
                    denom += 1.0

                # South
                if np.isfinite(var[jm1, i]):
                    numer += var[jm1, i]
                    denom += 1.0

                # self
                if np.isfinite(var[j, i]):
                    numer += denom * var[j, i]
                    denom *= 2.0

                if denom > 0.0:
                    work[j, i] = numer / denom
                    if np.isnan(var[j, i]):
                        done = False
                    else:
                        delta = np.fabs(var[j, i] - work[j, i])
                        if delta > tol * np.abs(var[j, i]):
                            done = False

        var[1:nlat, :] = work[1:nlat, :]
        if iter_cnt > max_iter:
            done = True


@jit(nopython=True, parallel=True)
def _iterative_fill_sor(nlat, nlon, var, fillmask, tol, rc, max_iter, ltripole):
    """Iterative land fill algorithm via SOR solution of Laplace Equation."""

    # Compute a zonal mean to use as a first guess
    # Apprarently jit doesn't like masked arrays so loop it out

    zoncnt = np.zeros(nlat)
    zonavg = np.zeros(nlat)
    for j in range(0, nlat):
        zoncnt[j] = np.sum(np.where(fillmask[j, :], 0, 1))
        zonavg[j] = np.sum(np.where(fillmask[j, :], 0, var[j, :]))
        if zoncnt[j] != 0:
            zonavg[j] = zonavg[j] / zoncnt[j]

    # Fill missing zonal averages for rows that are entirely land
    for j in range(0, nlat - 1):  # northward pass
        if zoncnt[j] > 0 and zoncnt[j + 1] == 0:
            zoncnt[j + 1] = 1
            zonavg[j + 1] = zonavg[j]
    for j in range(nlat - 1, 0, -1):  # southward pass
        if zoncnt[j] > 0 and zoncnt[j - 1] == 0:
            zoncnt[j - 1] = 1
            zonavg[j - 1] = zonavg[j]

    # Replace the input array missing values with zonal average as first guess
    for j in range(0, nlat):
        for i in range(0, nlon):
            if fillmask[j, i]:
                var[j, i] = zonavg[j]

    # Now do the iterative 2D fill
    res = np.zeros((nlat, nlon))  # work array hold residuals
    res_max = tol
    iter_cnt = 0
    while iter_cnt < max_iter and res_max >= tol:
        res[:] = 0.0  # reset the residual to zero for this iteration

        # assume bottom row is all land, leave it set to zonal average
        # deal with top row separately below
        for j in range(1, nlat - 1):
            jm1 = j - 1
            jp1 = j + 1

            for i in range(0, nlon):
                if fillmask[j, i]:
                    im1 = (i - 1) % nlon  # assume periodic in x
                    ip1 = (i + 1) % nlon

                    # this is SOR
                    res[j, i] = (
                        var[j, ip1] + var[j, im1] + var[jm1, i] + var[jp1, i] - 4.0 * var[j, i]
                    )
                    var[j, i] = var[j, i] + rc * 0.25 * res[j, i]

        # do the top row if there was some valid data there in the input
        # otherwise leave it set to zonal average of northernmost row with valid data
        if zoncnt[nlat - 1] > 1:
            j = nlat - 1
            jm1 = j - 1
            jp1 = j
            for i in range(0, nlon):
                if fillmask[j, i]:
                    im1 = (i - 1) % nlon
                    ip1 = (i + 1) % nlon
                    io = nlon - 1 - i  # tripole b.c

                    if ltripole:  # use cross-pole periodicity
                        res[j, i] = (
                            var[j, ip1] + var[j, im1] + var[jp1, io] + var[jm1, i] - 4.0 * var[j, i]
                        )
                        var[j, i] = var[j, i] + rc * 0.25 * res[j, i]
                    else:  # do a 1D smooth on pole row
                        res[j, i] = var[j, ip1] + var[j, im1] - 2.0 * var[j, i]
                        var[j, i] = var[j, i] + rc * 0.5 * res[j, i]

        res_max = np.max(np.fabs(res)) / np.max(np.fabs(var))
        iter_cnt += 1



import numpy as np
import pkg_resources
import xarray as xr
import yaml

from .grid import get_grid, grid_defs

region_def_file = pkg_resources.resource_filename('pop_tools', 'region_mask_definitions.yaml')
# open defined region masks
with open(region_def_file) as f:
    all_region_defs = yaml.safe_load(f)


def region_mask_3d(grid_name, mask_name=None, region_defs=None):
    """Generate as 3-D region mask with dimensions ('region', 'nlat', 'nlon').

    Parameters
    ----------

    grid_name : str
        Name of grid (i.e., POP_gx3v7, POP_gx1v7, POP_tx0.1v3)

    mask_name : str, optional
        Name of the pre-defined mask. If not provided and `region_defs` is
        not provided, the default `REGION_MASK` is used.

    region_defs : dict, optional
        Dictionary containing region definitions. The dictionary has the
        following form:

            region_defs = {region1_name: list_of_criteria_dicts_1, region2_name: list_of_criteria_dicts_2,...}

        `list_of_criteria_dicts` is a list of dictionaries; each must include
        the keys 'match' or 'bounds'. For instance:

            list_of_criteria_dicts = [{'match': {'REGION_MASK': [1, 2, 3, 6]}, 'bounds': {'TLAT': [-90., -30.]}}]

        will identify a region where the default POP grid `REGION_MASK` values match
        the specified list and `TLAT` is between 90°S and 30°S. Multiple entries
        in the `list_of_criteria_dicts` are applied with an "or" condition.

    Returns
    -------

    mask3d : xarray.DataArray
        Region mask DataArray with dimensions ('region', 'nlat', 'nlon').
        `mask3d` has "ones" within a region and "zeroes" outside it. A `region`
        coordinate is included that contains the region names.
    """
    ds = get_grid(grid_name)
    nlat, nlon = ds.REGION_MASK.shape

    if region_defs is None and mask_name is None:
        mask_name = 'default'

    elif region_defs is not None and mask_name is None:
        mask_name = 'user-defined'

    if region_defs is None:
        region_defs = _get_region_definitions(ds, grid_name, mask_name)

    _verify_region_def_schema(region_defs)

    region_names = list(region_defs.keys())
    region_coord = xr.DataArray(region_names, dims=('region'))

    mask_shape = region_coord.shape + (nlat, nlon)
    mask_dims = region_coord.dims + ds.REGION_MASK.dims
    mask3d = xr.DataArray(
        np.zeros(mask_shape),
        dims=mask_dims,
        coords={'region': region_coord},
        attrs={'mask_name': mask_name},
    )

    not_land = np.where(ds.KMT > 0, 1, 0)

    # loop over region names
    for i, (region_name, crit_list) in enumerate(region_defs.items()):

        # loop over list of criteria groups
        mask = np.zeros((nlat, nlon), dtype=bool)
        for crit in crit_list:

            mask_and = np.ones((nlat, nlon), dtype=bool)

            # loop over each criterion
            for crit_type, crit_dict in crit.items():

                # apply bounds or match
                if crit_type == 'bounds':
                    for field, bounds in crit_dict.items():
                        if bounds[0] > bounds[1]:
                            mask_and = (bounds[0] <= ds[field]) | (
                                ds[field] <= bounds[1]
                            ) & mask_and
                        else:
                            mask_and = (
                                (bounds[0] <= ds[field]) & (ds[field] <= bounds[1]) & mask_and
                            )

                elif crit_type == 'match':
                    mask_or = np.zeros((nlat, nlon), dtype=bool)
                    for field, value in crit_dict.items():
                        for value_i in value:
                            mask_or = (ds[field] == value_i) | mask_or
                    mask_and = mask_or & mask_and

                else:
                    raise ValueError(f'unknown criteria type: {crit_type}')

            # update mask logic
            mask = mask | mask_and

        # make mask
        mask3d[i, :, :] = xr.where(mask & not_land, 1.0, 0.0)

    return mask3d


def _verify_region_def_schema(region_defs):
    """Ensure region_defs dictionary conforms to expected schema."""
    # loop over region names
    for i, (region_name, crit_list) in enumerate(region_defs.items()):
        assert isinstance(crit_list, list)
        # loop over list of criteria groups
        for crit in crit_list:
            assert isinstance(crit, dict)
            # loop over each criterion
            for crit_type, crit_dict in crit.items():
                assert crit_type in ['bounds', 'match']


def _get_region_definitions(ds, grid_name, mask_name):
    """Return the region mask definition from `region_mask_definitions.yaml`."""
    if mask_name == 'default':
        grid_attrs = grid_defs[grid_name]

        regions = np.unique(ds.REGION_MASK)
        regions = regions[regions != 0]

        if 'region_mask_regions' in grid_attrs:
            region_names = list(grid_attrs['region_mask_regions'].keys())
            region_index = list(grid_attrs['region_mask_regions'].values())

        else:
            region_names = regions
            region_index = regions

        return {k: [{'match': {'REGION_MASK': [v]}}] for k, v in zip(region_names, region_index)}
    else:
        if mask_name not in all_region_defs[grid_name]:
            raise ValueError(
                f'''unknown region mask: {mask_name}
                             the following regions masks are defined: {list_region_masks(grid_name)}
                             '''
            )

        return all_region_defs[grid_name][mask_name]


def list_region_masks(grid_name):
    """Get a list of the pre-defined region masks.

    Parameters
    ----------

    grid_name : str
        Name of grid (i.e., POP_gx3v7, POP_gx1v7, POP_tx0.1v3)

    Returns
    -------

    region_mask_list : list
        List of defined region mask names.
    """
    if grid_name in all_region_defs:
        return list(all_region_defs[grid_name].keys())
    else:
        return []



"""
Functions to load sample data
"""

import os
from pathlib import Path

import pkg_resources
import pooch

DATASETS = pooch.create(
    path=['~', '.pop_tools', 'data'],
    version_dev='master',
    base_url='ftp://ftp.cgd.ucar.edu/archive/aletheia-data/cesm-data/ocn/',
    env='POP_TOOLS_DATA_DIR',
)
DATASETS.load_registry(pkg_resources.resource_stream('pop_tools', 'data_registry.txt'))


class UnzipZarr(pooch.processors.Unzip):
    """
    Processor that unpacks a zarr store zip archive and
    returns the zarr store path.
    """

    def __call__(self, fname, action, pooch):
        """
        Extract all files from the given archive.
        Parameters
        ----------
        fname : str
            Full path of the zipped file in local storage.
        action : str
            Indicates what action was taken by :meth:`pooch.Pooch.fetch`:
            * ``"download"``: File didn't exist locally and was downloaded
            * ``"update"``: Local file was outdated and was re-download
            * ``"fetch"``: File exists and is updated so it wasn't downloaded
        pooch : :class:`pooch.Pooch`
            The instance of :class:`pooch.Pooch` that is calling this.
        Returns
        -------
        zarr_store : str
            A full path to a zarr store in the extracted archive.
        """
        extract_dir = fname + self.suffix
        if action in ('update', 'download') or not os.path.exists(extract_dir):
            # Make sure that the folder with the extracted files exists
            if not os.path.exists(extract_dir):
                os.makedirs(extract_dir)
            self._extract_file(fname, extract_dir)
        # Get a list of all file names (including subdirectories) in our folder
        # of unzipped files.
        fnames = [
            os.path.join(path, fname) for path, _, files in os.walk(extract_dir) for fname in files
        ]
        # Return the path of the zarr store
        return Path(sorted(fnames)[0]).parent.as_posix()



"""
Module for modifying pop model output to be compatible with xgcm

# Ref: https://gist.github.com/rabernat/933bc785c99828352f343e48d0da6e22
"""

import numpy as np
import xarray as xr


def _add_pop_dims_to_dataset(ds):
    ds_new = ds.copy()
    ds_new['nlon_u'] = xr.Variable(
        ('nlon_u'), np.arange(len(ds.nlon)) + 1, {'axis': 'X', 'c_grid_axis_shift': 0.5}
    )
    ds_new['nlat_u'] = xr.Variable(
        ('nlat_u'), np.arange(len(ds.nlat)) + 1, {'axis': 'Y', 'c_grid_axis_shift': 0.5}
    )
    ds_new['nlon_t'] = xr.Variable(('nlon_t'), np.arange(len(ds.nlon)) + 0.5, {'axis': 'X'})
    ds_new['nlat_t'] = xr.Variable(('nlat_t'), np.arange(len(ds.nlat)) + 0.5, {'axis': 'Y'})

    # add metadata to z grid
    if 'z_t' in ds_new.variables:
        ds_new['z_t'].attrs.update({'axis': 'Z'})
    if 'z_w' in ds_new.variables:
        ds_new['z_w'].attrs.update({'axis': 'Z', 'c_grid_axis_shift': -0.5})
    if 'z_w_top' in ds_new.variables:
        ds_new['z_w_top'].attrs.update({'axis': 'Z', 'c_grid_axis_shift': -0.5})
    if 'z_w_bot' in ds_new.variables:
        ds_new['z_w_bot'].attrs.update({'axis': 'Z', 'c_grid_axis_shift': 0.5})

    return ds_new


def _dims_from_grid_loc(grid_loc):
    grid_loc = str(grid_loc)
    ndim = int(grid_loc[0])
    x_loc_key = int(grid_loc[1])
    y_loc_key = int(grid_loc[2])
    z_loc_key = int(grid_loc[3])

    x_loc = {1: 'nlon_t', 2: 'nlon_u', 3: 'nlon_u'}[x_loc_key]
    y_loc = {1: 'nlat_t', 2: 'nlat_u', 3: 'nlat_t'}[y_loc_key]
    z_loc = {0: 'surface', 1: 'z_t', 2: 'z_w', 3: 'z_w_bot', 4: 'z_t_150m'}[z_loc_key]

    if ndim == 3:
        return z_loc, y_loc, x_loc
    elif ndim == 2:
        return y_loc, x_loc


def _label_coord_grid_locs(ds):
    grid_locs = {
        'ANGLE': '2220',
        'ANGLET': '2110',
        'DXT': '2110',
        'DXU': '2220',
        'DYT': '2110',
        'DYU': '2220',
        'DZU': '3221',
        'DZT': '3111',
        'HT': '2110',
        'HU': '2220',
        'HTE': '2210',
        'HTN': '2120',
        'HUS': '2210',
        'HUW': '2120',
        'KMT': '2110',
        'KMU': '2220',
        'REGION_MASK': '2110',
        'TAREA': '2110',
        'TLAT': '2110',
        'TLONG': '2110',
        'UAREA': '2220',
        'ULAT': '2220',
        'ULONG': '2220',
    }
    ds_new = ds.copy()
    for vname, grid_loc in grid_locs.items():
        if vname in ds.variables:
            ds_new[vname].attrs['grid_loc'] = grid_loc
    return ds_new


def relabel_pop_dims(ds):
    """Return a new xarray dataset with distinct dimensions for variables at different
    grid points.
    """
    ds_new = _label_coord_grid_locs(ds)
    ds_new = _add_pop_dims_to_dataset(ds_new)
    for vname in ds_new.variables:
        if 'grid_loc' in ds_new[vname].attrs:
            da = ds_new[vname]
            dims_orig = da.dims
            new_spatial_dims = _dims_from_grid_loc(da.attrs['grid_loc'])
            if dims_orig[0] == 'time':
                dims = ('time',) + new_spatial_dims
            else:
                dims = new_spatial_dims
            ds_new[vname] = xr.Variable(dims, da.data, da.attrs, da.encoding, fastpath=True)
    old_coords = ['nlat', 'nlon']
    for coord in old_coords:
        if coord in ds_new.coords:
            ds_new = ds_new.drop_vars(coord)
    if 'z_w_top' in ds_new.dims and 'z_w' in ds_new.dims:
        ds_new = ds_new.drop('z_w_top').rename({'z_w': 'z_w_top'})
    return ds_new


def to_xgcm_grid_dataset(ds, **kwargs):
    """Modify POP model output to be compatible with xgcm.

    Parameters
    ----------
    ds : xarray.Dataset
        An xarray Dataset
    kwargs:
       Additional keyword arguments are passed through to `xgcm.Grid` class.

    Returns
    -------
    grid : xgcm.Grid
        An xgcm Grid object
    ds_new : xarray.Dataset
        Xarray dataset with distinct dimensions for variables at different
        grid points.

    Examples
    --------
    >>> from pop_tools import DATASETS
    >>> import xarray as xr
    >>> filepath = DATASETS.fetch("g.e20.G.TL319_t13.control.001_hfreq-coarsen.nc")
    >>> ds = xr.open_dataset(filepath)
    >>> ds
    <xarray.Dataset>
    Dimensions:          (d2: 2, nlat: 240, nlon: 360, time: 1, z_t: 62, z_t_150m: 15, z_w: 62, z_w_top: 62)
    Coordinates:
    * time             (time) object 0050-01-01 00:00:00
    * z_t              (z_t) float32 500.0 1500.0 2500.0 ... 562499.06 587499.06
    * z_t_150m         (z_t_150m) float32 500.0 1500.0 2500.0 ... 13500.0 14500.0
    * z_w              (z_w) float32 0.0 1000.0 2000.0 ... 549999.06 574999.06
    * z_w_top          (z_w_top) float32 0.0 1000.0 2000.0 ... 549999.06 574999.06
        TLONG            (nlat, nlon) float64 ...
        TLAT             (nlat, nlon) float64 ...
    Dimensions without coordinates: d2, nlat, nlon
    Data variables:
        dz               (z_t) float32 ...
        dzw              (z_w) float32 ...
        KMT              (nlat, nlon) float64 ...
        REGION_MASK      (nlat, nlon) float64 ...
        TAREA            (nlat, nlon) float64 ...
        DXU              (nlat, nlon) float64 ...
        DYU              (nlat, nlon) float64 ...
        DYT              (nlat, nlon) float64 ...
        time_bound       (time, d2) object ...
        UEU              (time, z_t, nlat, nlon) float32 ...
        VNU              (time, z_t, nlat, nlon) float32 ...
        S_FLUX_ROFF_VSF  (time, z_w_top, nlat, nlon) float32 ...
    >>> grid, ds_new = to_xgcm_grid_dataset(ds)
    >>> grid
    <xgcm.Grid>
    X Axis (periodic):
        * center   nlon_t --> right
        * right    nlon_u --> center
    Z Axis (periodic):
        * center   z_t --> left
        * left     z_w_top --> center
    Y Axis (periodic):
        * center   nlat_t --> right
        * right    nlat_u --> center
    >>> ds_new
    <xarray.Dataset>
    Dimensions:          (d2: 2, nlat_t: 240, nlat_u: 240, nlon_t: 360, nlon_u: 360, time: 1, z_t: 62, z_t_150m: 15, z_w: 62, z_w_top: 62)
    Coordinates:
    * time             (time) object 0050-01-01 00:00:00
    * z_t              (z_t) float32 500.0 1500.0 2500.0 ... 562499.06 587499.06
    * z_t_150m         (z_t_150m) float32 500.0 1500.0 2500.0 ... 13500.0 14500.0
    * z_w              (z_w) float32 0.0 1000.0 2000.0 ... 549999.06 574999.06
    * z_w_top          (z_w_top) float32 0.0 1000.0 2000.0 ... 549999.06 574999.06
      TLONG            (nlat_t, nlon_t) float64 nan nan nan nan ... nan nan nan
      TLAT             (nlat_t, nlon_t) float64 nan nan nan nan ... nan nan nan
    * nlon_u           (nlon_u) int64 1 2 3 4 5 6 7 ... 355 356 357 358 359 360
    * nlat_u           (nlat_u) int64 1 2 3 4 5 6 7 ... 235 236 237 238 239 240
    * nlon_t           (nlon_t) float64 0.5 1.5 2.5 3.5 ... 357.5 358.5 359.5
    * nlat_t           (nlat_t) float64 0.5 1.5 2.5 3.5 ... 237.5 238.5 239.5
    Dimensions without coordinates: d2
    Data variables:
        dz               (z_t) float32 ...
        dzw              (z_w) float32 ...
        KMT              (nlat_t, nlon_t) float64 nan nan nan nan ... nan nan nan
        REGION_MASK      (nlat_t, nlon_t) float64 nan nan nan nan ... nan nan nan
        TAREA            (nlat_t, nlon_t) float64 nan nan nan nan ... nan nan nan
        DXU              (nlat_u, nlon_u) float64 nan nan nan nan ... nan nan nan
        DYU              (nlat_u, nlon_u) float64 nan nan nan nan ... nan nan nan
        DYT              (nlat_t, nlon_t) float64 nan nan nan nan ... nan nan nan
        time_bound       (time, d2) object ...
        UEU              (time, z_t, nlat_u, nlon_u) float32 inf inf inf ... inf inf
        VNU              (time, z_t, nlat_t, nlon_u) float32 inf inf inf ... inf inf
        S_FLUX_ROFF_VSF  (time, z_w, nlat_t, nlon_t) float32 inf inf inf ... inf inf

    """

    try:
        import xgcm
    except ImportError:
        raise ImportError(
            """to_xgcm_grid_dataset() function requires the `xgcm` package. \nYou can install it via PyPI or Conda"""
        )
    ds_new = relabel_pop_dims(ds)
    grid = xgcm.Grid(ds_new, **kwargs)
    return grid, ds_new



import os
from pathlib import Path

import numpy as np
import pkg_resources
import pooch
import xarray as xr
import yaml
from numba import jit, prange

try:
    from tqdm import tqdm
except ImportError:
    tqdm = None


# On Cheyenne/Casper and/or CGD machines, use local inputdata directory
# See: https://github.com/NCAR/pop-tools/issues/24#issue-523701065
# The name of the environment variable that can overwrite the path argument
cesm_data_root_path = os.environ.get('CESMDATAROOT')

if cesm_data_root_path is not None and os.path.exists(cesm_data_root_path):
    INPUTDATA_DIR = cesm_data_root_path
else:
    # This is still the default in case the environment variable isn't defined
    INPUTDATA_DIR = ['~', '.pop_tools']


INPUTDATA = pooch.create(
    path=INPUTDATA_DIR,
    version_dev='master',
    base_url='https://svn-ccsm-inputdata.cgd.ucar.edu/trunk/',
)


INPUTDATA.load_registry(pkg_resources.resource_stream('pop_tools', 'inputdata_registry.txt'))

if tqdm is not None:
    downloader = pooch.HTTPDownloader(progressbar=True, verify=False, allow_redirects=True)
else:
    downloader = pooch.HTTPDownloader(verify=False, allow_redirects=True)


grid_def_file = pkg_resources.resource_filename('pop_tools', 'pop_grid_definitions.yaml')
input_templates_dir = pkg_resources.resource_filename('pop_tools', 'input_templates')

with open(grid_def_file) as f:
    grid_defs = yaml.safe_load(f)


def fetch(self, fname, processor=None, downloader=None):

    """
    This is a modified version of Pooch.fetch() method. This modification is necessary
    due to the fact that on Cheyenne/Casper path to the local data storage folder points
    to a folder (CESMDATAROOT: /glade/p/cesmdata/cseg), and this is not a location that
    we have permissions to write to.

    Parameters
    ----------
    fname : str
        The file name (relative to the *base_url* of the remote data
        storage) to fetch from the local storage.
    processor : None or callable
        If not None, then a function (or callable object) that will be
        called before returning the full path and after the file has been
        downloaded (if required).
    downloader : None or callable
        If not None, then a function (or callable object) that will be
        called to download a given URL to a provided local file name. By
        default, downloads are done through HTTP without authentication
        using :class:`pooch.HTTPDownloader`.
    Returns
    -------
    full_path : str
        The absolute path (including the file name) of the file in the
        local storage.

    """

    self._assert_file_in_registry(fname)
    url = self.get_url(fname)
    full_path = self.abspath / fname
    known_hash = self.registry[fname]
    abspath = str(self.abspath)
    action, verb = pooch.core.download_action(full_path, known_hash)

    if action in ('download', 'update'):
        pooch.utils.get_logger().info("%s file '%s' from '%s' to '%s'.", verb, fname, url, abspath)
        if downloader is None:
            downloader = pooch.downloaders.choose_downloader(url)

        pooch.core.stream_download(url, full_path, known_hash, downloader, pooch=self)

    if processor is not None:
        return processor(str(full_path), action, self)

    return str(full_path)


# Override fetch method at instance level
# Reference: https://stackoverflow.com/a/46757134/7137180
# Replace fetch() with modified fetch() for this object only
INPUTDATA.fetch = fetch.__get__(INPUTDATA, pooch.Pooch)


def get_grid(grid_name, scrip=False):
    """Return a xarray.Dataset() with POP grid variables.

    Parameters
    ----------

    grid_name : str
      Name of grid (i.e., POP_gx3v7, POP_gx1v7, POP_tx0.1v3)

    scrip : boolean, optional
      Return grid in SCRIP format

    Returns
    -------

    dso : `xarray.Dataset`
      Dataset containing POP grid variables.
    """

    if grid_name not in grid_defs:
        raise ValueError(
            f"""Unknown grid: {grid_name}
             Please select from the following: {list(grid_defs.keys())}"""
        )

    grid_attrs = grid_defs[grid_name]

    nlat = grid_attrs['lateral_dims'][0]
    nlon = grid_attrs['lateral_dims'][1]

    # read horizontal grid
    horiz_grid_fname = INPUTDATA.fetch(grid_attrs['horiz_grid_fname'], downloader=downloader)
    grid_file_data = np.fromfile(horiz_grid_fname, dtype='>f8', count=-1)
    grid_file_data = grid_file_data.reshape((7, nlat, nlon))

    ULAT = grid_file_data[0, :, :].astype(np.float)
    ULONG = grid_file_data[1, :, :].astype(np.float)
    HTN = grid_file_data[2, :, :].astype(np.float)
    HTE = grid_file_data[3, :, :].astype(np.float)

    # compute TLAT, TLONG
    TLAT = np.empty((nlat, nlon), dtype=np.float)
    TLONG = np.empty((nlat, nlon), dtype=np.float)
    _compute_TLAT_TLONG(ULAT, ULONG, TLAT, TLONG, nlat, nlon)

    # generate DXT, DYT
    # DXT[i,j] = (HTN[i,j] + HTN[i,j−1])/2
    DXT = np.empty((nlat, nlon))
    DXT[1:, :] = 0.5 * (HTN[: nlat - 1, :] + HTN[1:, :])
    # DXT[0, :] = 0.5 * (2 * HTN[0, :] - HTN[1, :] + HTN[0, :])
    DXT[0, :] = 0.5 * (HTN[0, :] + HTN[nlat - 1, :])

    # DYT[i,j] = (HTE[i,j] + HTE[i−1,j])/2
    DYT = np.empty((nlat, nlon))
    DYT[:, 1:] = 0.5 * (HTE[:, : nlon - 1] + HTE[:, 1:])
    DYT[:, 0] = 0.5 * (HTE[:, nlon - 1] + HTE[:, 0])

    # generate DXU, DYU
    # DXU[i,j] = (HTN[i,j] + HTN[i+1,j])/2
    DXU = np.empty((nlat, nlon))
    DXU[:, : nlon - 1] = 0.5 * (HTN[:, : nlon - 1] + HTN[:, 1:])
    DXU[:, nlon - 1] = 0.5 * (HTN[:, nlon - 1] + HTN[:, 0])

    # DYU[i,j] = (HTE[i,j] + HTE[i,j+1])/2
    DYU = np.empty((nlat, nlon))
    DYU[: nlat - 1, :] = 0.5 * (HTE[: nlat - 1, :] + HTE[1:, :])
    DYU[nlat - 1, :] = 0.5 * (HTE[nlat - 1, :] + HTE[0, :])

    # compute TAREA, UAREA
    TAREA = DXT * DYT
    UAREA = DXU * DYU

    # vertical grid
    vert_grid_fname = os.path.join(input_templates_dir, grid_attrs['vert_grid_file'])
    tmp = np.loadtxt(vert_grid_fname)
    dz = tmp[:, 0]
    depth_edges = np.concatenate(([0.0], np.cumsum(dz)))
    z_w = depth_edges[0:-1]
    z_w_bot = depth_edges[1:]
    z_t = depth_edges[0:-1] + 0.5 * dz

    # read KMT
    topography_fname = INPUTDATA.fetch(grid_attrs['topography_fname'], downloader=downloader)
    kmt_flat = np.fromfile(topography_fname, dtype='>i4', count=-1)
    assert kmt_flat.shape[0] == (
        nlat * nlon
    ), f'unexpected dims in topography file: {grid_attrs["topography_fname"]}'
    assert kmt_flat.max() <= len(z_t), 'Max KMT > length z_t'
    KMT = kmt_flat.reshape(grid_attrs['lateral_dims']).astype(np.int32)

    # read REGION_MASK
    region_mask_fname = INPUTDATA.fetch(grid_attrs['region_mask_fname'], downloader=downloader)
    region_mask_flat = np.fromfile(region_mask_fname, dtype='>i4', count=-1)
    assert region_mask_flat.shape[0] == (
        nlat * nlon
    ), f'unexpected dims in region_mask file: {grid_attrs["region_mask_fname"]}'
    REGION_MASK = region_mask_flat.reshape(grid_attrs['lateral_dims']).astype(np.int32)

    # output dataset
    dso = xr.Dataset()
    if scrip:
        corner_lat, corner_lon = _compute_corners(ULAT, ULONG)

        dso['grid_dims'] = xr.DataArray(np.array([nlon, nlat], dtype=np.int32), dims=('grid_rank',))
        dso.grid_dims.encoding = {'dtype': np.int32, '_FillValue': None}

        dso['grid_center_lat'] = xr.DataArray(
            np.rad2deg(TLAT.reshape((-1,))), dims=('grid_size'), attrs={'units': 'degrees'}
        )
        dso.grid_center_lat.encoding = {'dtype': np.float64, '_FillValue': None}

        dso['grid_center_lon'] = xr.DataArray(
            np.rad2deg(TLONG.reshape((-1,))), dims=('grid_size'), attrs={'units': 'degrees'}
        )
        dso.grid_center_lon.encoding = {'dtype': np.float64, '_FillValue': None}

        dso['grid_corner_lat'] = xr.DataArray(
            np.rad2deg(corner_lat.reshape((-1, 4))),
            dims=('grid_size', 'grid_corners'),
            attrs={'units': 'degrees'},
        )
        dso.grid_corner_lat.encoding = {'dtype': np.float64, '_FillValue': None}

        dso['grid_corner_lon'] = xr.DataArray(
            np.rad2deg(corner_lon.reshape((-1, 4))),
            dims=('grid_size', 'grid_corners'),
            attrs={'units': 'degrees'},
        )
        dso.grid_corner_lon.encoding = {'dtype': np.float64, '_FillValue': None}

        dso['grid_imask'] = xr.DataArray(
            np.where(KMT > 0, 1, 0).reshape((-1,)), dims=('grid_size'), attrs={'units': 'unitless'}
        )
        dso.grid_imask.encoding = {'dtype': np.int32, '_FillValue': None}

        grid_attrs.update({'conventions': 'SCRIP'})

    else:
        TLONG = np.where(TLONG < 0.0, TLONG + 2 * np.pi, TLONG)

        dso['TLAT'] = xr.DataArray(
            np.rad2deg(TLAT),
            dims=('nlat', 'nlon'),
            attrs={'units': 'degrees_north', 'long_name': 'T-grid latitude'},
        )

        dso['TLONG'] = xr.DataArray(
            np.rad2deg(TLONG),
            dims=('nlat', 'nlon'),
            attrs={'units': 'degrees_east', 'long_name': 'T-grid longitude'},
        )

        dso['ULAT'] = xr.DataArray(
            np.rad2deg(ULAT),
            dims=('nlat', 'nlon'),
            attrs={'units': 'degrees_north', 'long_name': 'U-grid latitude'},
        )

        dso['ULONG'] = xr.DataArray(
            np.rad2deg(ULONG),
            dims=('nlat', 'nlon'),
            attrs={'units': 'degrees_east', 'long_name': 'U-grid longitude'},
        )

        dso['DXT'] = xr.DataArray(
            DXT,
            dims=('nlat', 'nlon'),
            attrs={
                'units': 'cm',
                'long_name': 'x-spacing centered at T points',
                'coordinates': 'TLONG TLAT',
            },
        )

        dso['DYT'] = xr.DataArray(
            DYT,
            dims=('nlat', 'nlon'),
            attrs={
                'units': 'cm',
                'long_name': 'y-spacing centered at T points',
                'coordinates': 'TLONG TLAT',
            },
        )

        dso['DXU'] = xr.DataArray(
            DXU,
            dims=('nlat', 'nlon'),
            attrs={
                'units': 'cm',
                'long_name': 'x-spacing centered at U points',
                'coordinates': 'ULONG ULAT',
            },
        )

        dso['DYU'] = xr.DataArray(
            DYU,
            dims=('nlat', 'nlon'),
            attrs={
                'units': 'cm',
                'long_name': 'y-spacing centered at U points',
                'coordinates': 'ULONG ULAT',
            },
        )

        dso['TAREA'] = xr.DataArray(
            TAREA,
            dims=('nlat', 'nlon'),
            attrs={'units': 'cm^2', 'long_name': 'area of T cells', 'coordinates': 'TLONG TLAT'},
        )

        dso['UAREA'] = xr.DataArray(
            UAREA,
            dims=('nlat', 'nlon'),
            attrs={'units': 'cm^2', 'long_name': 'area of U cells', 'coordinates': 'ULONG ULAT'},
        )

        dso['KMT'] = xr.DataArray(
            KMT,
            dims=('nlat', 'nlon'),
            attrs={
                'long_name': 'k Index of Deepest Grid Cell on T Grid',
                'coordinates': 'TLONG TLAT',
            },
        )

        dso['REGION_MASK'] = xr.DataArray(
            REGION_MASK,
            dims=('nlat', 'nlon'),
            attrs={
                'long_name': 'basin index number (signed integers)',
                'coordinates': 'TLONG TLAT',
            },
        )

        dso['z_t'] = xr.DataArray(
            z_t,
            dims=('z_t'),
            name='z_t',
            attrs={
                'units': 'cm',
                'long_name': 'depth from surface to midpoint of layer',
                'positive': 'down',
            },
        )

        dso['dz'] = xr.DataArray(
            dz,
            dims=('z_t'),
            coords={'z_t': dso.z_t},
            attrs={'units': 'cm', 'long_name': 'thickness of layer k'},
        )

        dso['z_w'] = xr.DataArray(
            z_w,
            dims=('z_w'),
            attrs={
                'units': 'cm',
                'positive': 'down',
                'long_name': 'depth from surface to top of layer',
            },
        )

        dso['z_w_bot'] = xr.DataArray(
            z_w_bot,
            dims=('z_w_bot'),
            attrs={
                'units': 'cm',
                'positive': 'down',
                'long_name': 'depth from surface to bottom of layer',
            },
        )

    grid_attrs.update({'title': f'{grid_name} grid'})

    dso.attrs = grid_attrs

    # Remove region_mask_regions
    if 'region_mask_regions' in dso.attrs:
        regions = dso.attrs.pop('region_mask_regions')
        region_names, region_vals = list(zip(*regions.items()))
        region_coord = list(range(len(regions)))
        dso['region_name'] = xr.DataArray(list(region_names), coords=[region_coord], dims=['nreg'])
        dso['region_val'] = xr.DataArray(list(region_vals), coords=[region_coord], dims=['nreg'])
        dso['region_val'].attrs['coordinate'] = 'region_name'

    return dso


@jit(nopython=True, parallel=True)
def _compute_TLAT_TLONG(ULAT, ULONG, TLAT, TLONG, nlat, nlon):
    """Compute TLAT and TLONG from ULAT, ULONG"""

    for j in prange(1, nlat):
        jm1 = j - 1
        for i in prange(0, nlon):
            im1 = np.mod(i - 1 + nlon, nlon)

            tmp = np.cos(ULAT[jm1, im1])
            xsw = np.cos(ULONG[jm1, im1]) * tmp
            ysw = np.sin(ULONG[jm1, im1]) * tmp
            zsw = np.sin(ULAT[jm1, im1])

            tmp = np.cos(ULAT[jm1, i])
            xse = np.cos(ULONG[jm1, i]) * tmp
            yse = np.sin(ULONG[jm1, i]) * tmp
            zse = np.sin(ULAT[jm1, i])

            tmp = np.cos(ULAT[j, im1])
            xnw = np.cos(ULONG[j, im1]) * tmp
            ynw = np.sin(ULONG[j, im1]) * tmp
            znw = np.sin(ULAT[j, im1])

            tmp = np.cos(ULAT[j, i])
            xne = np.cos(ULONG[j, i]) * tmp
            yne = np.sin(ULONG[j, i]) * tmp
            zne = np.sin(ULAT[j, i])

            xc = 0.25 * (xsw + xse + xnw + xne)
            yc = 0.25 * (ysw + yse + ynw + yne)
            zc = 0.25 * (zsw + zse + znw + zne)

            r = np.sqrt(xc * xc + yc * yc + zc * zc)

            TLAT[j, i] = np.arcsin(zc / r)
            TLONG[j, i] = np.arctan2(yc, xc)

    # generate bottom row
    TLAT[0, :] = TLAT[1, :] - (TLAT[2, :] - TLAT[1, :])
    TLONG[0, :] = TLONG[1, :] - (TLONG[2, :] - TLONG[1, :])


def _compute_corners(ULAT, ULONG):
    """Compute grid corners."""

    nlat, nlon = ULAT.shape
    corner_lat = np.empty((nlat, nlon, 4), dtype=np.float)
    corner_lon = np.empty((nlat, nlon, 4), dtype=np.float)

    # NE corner
    corner_lat[:, :, 0] = ULAT
    corner_lon[:, :, 0] = ULONG

    # NW corner (copy from NE corner of column to the left, assume zonal periodic bc)
    corner_lat[:, :, 1] = np.roll(corner_lat[:, :, 0], 1, axis=1)
    corner_lon[:, :, 1] = np.roll(corner_lon[:, :, 0], 1, axis=1)

    # SW corner (copy from NW corner of row below, bottom row is extrapolated from 2 rows above)
    corner_lat[1:nlat, :, 2] = corner_lat[0 : nlat - 1, :, 1]
    corner_lon[1:nlat, :, 2] = corner_lon[0 : nlat - 1, :, 1]
    corner_lat[0, :, 2] = corner_lat[1, :, 2] - (corner_lat[2, :, 2] - corner_lat[1, :, 2])
    corner_lon[0, :, 2] = corner_lon[1, :, 2] - (corner_lon[2, :, 2] - corner_lon[1, :, 2])

    # SE corner (copy from NE corner of row below, bottom row is extrapolated from 2 rows above)
    corner_lat[1:nlat, :, 3] = corner_lat[0 : nlat - 1, :, 0]
    corner_lon[1:nlat, :, 3] = corner_lon[0 : nlat - 1, :, 0]
    corner_lat[0, :, 3] = corner_lat[1, :, 3] - (corner_lat[2, :, 3] - corner_lat[1, :, 3])
    corner_lon[0, :, 3] = corner_lon[1, :, 3] - (corner_lon[2, :, 3] - corner_lon[1, :, 3])

    return corner_lat, corner_lon



import dask
import numpy as np
import xarray as xr
from numba import vectorize


@vectorize(['float64(float64)', 'float32(float32)'], nopython=True)
def compute_pressure(depth):
    """
    Convert depth in meters to pressure in bars.

    Parameters
    ----------
    depth : float
      Depth in meters

    Returns
    -------
    pressure : float
      Pressure in dbar
    """
    return (
        0.059808 * (np.exp(-0.025 * depth) - 1.0) + 0.100766 * depth + 2.28405e-7 * (depth ** 2.0)
    )


def eos(salt, temp, return_coefs=False, **kwargs):
    """
    Compute density as a function of salinity, temperature, and
    depth (or pressure).

    McDougall, T.J., D.R. Jackett, D.G. Wright, and R. Feistel, 2003:
    Accurate and Computationally Efficient Algorithms for Potential
    Temperature and Density of Seawater. J. Atmos. Oceanic Technol., 20,
    730–741, _`https://doi.org/10.1175/1520-0426(2003)20<730:AACEAF>2.0.CO;2`.

    test value:
        rho = 1033.213387 kg/m^3;
        S = 35.0 PSU, theta = 20.0 C, pressure = 2000.0 dbars

    Parameters
    ----------
    salt : float
        salinity, psu
    temp : float
        potential temperature, degree C
    return_coefs : boolean, optional [default=False]
        Logical, if true function returns 2 additional arguments:
        dRHOdS and dRHOdT
    depth : float, optional
        depth in meters, if not provided, pressure must be provided.
    pressure : float, optional
        depth in dbar

    Returns
    -------
    rho : float
        density kg/m^3
    dRHOdS : float, optional
        Derivative of rho with respect to salinity.
    dRHOdT : float, optional
        Derivative of rho with respect to temperature.
    """

    depth = kwargs.pop('depth', None)
    pressure = kwargs.pop('pressure', None)

    if kwargs:
        raise ValueError(f'unknown arguments: {kwargs}')

    if depth is None and pressure is None:
        raise ValueError('either depth or pressure must be supplied')
    else:
        d_or_p = depth if pressure is None else pressure

    use_xarray = False
    if any(isinstance(arg, xr.DataArray) for arg in [salt, temp, d_or_p]):
        if not all(isinstance(arg, xr.DataArray) for arg in [salt, temp, d_or_p]):
            raise ValueError('cannot operate on mixed types')
        use_xarray = True

    # compute pressure
    if pressure is None:
        if use_xarray:
            pressure = xr.full_like(depth, fill_value=np.nan)
            pressure[:] = 10.0 * compute_pressure(depth.data)  # dbar
        else:
            pressure = 10.0 * compute_pressure(depth)  # dbar

    # enforce min/max values
    tmin = -2.0
    tmax = 999.0
    smin = 0.0
    smax = 999.0

    if use_xarray:
        temp = temp.clip(tmin, tmax)
        salt = salt.clip(smin, smax)

        salt, temp, pressure = xr.broadcast(salt, temp, pressure)
        if isinstance(salt.data, dask.array.Array):
            pressure = pressure.chunk(salt.chunks)

        if return_coefs:
            RHO, dRHOdS, dRHOdT = _compute_eos_coeffs(salt, temp, pressure)

            dRHOdS.name = 'dRHOdS'
            dRHOdS.attrs['units'] = 'kg/m^3/degC'
            dRHOdS.attrs['long_name'] = 'Haline contraction coefficient'

            dRHOdT.name = 'dRHOdT'
            dRHOdT.attrs['units'] = 'kg/m^3/degC'
            dRHOdT.attrs['long_name'] = 'Thermal expansion coefficient'

        else:
            RHO = xr.apply_ufunc(
                _compute_eos,
                salt,
                temp,
                pressure,
                dask='parallelized',
                output_dtypes=[salt.dtype],
            )

        RHO.name = 'density'
        RHO.attrs['units'] = 'kg/m^3'
        RHO.attrs['long_name'] = 'Density'

    else:
        temp = np.clip(temp, tmin, tmax)
        salt = np.clip(salt, smin, smax)

        if return_coefs:
            RHO, dRHOdS, dRHOdT = _compute_eos_coeffs(salt, temp, pressure)
        else:
            RHO = _compute_eos(salt, temp, pressure)

    if return_coefs:
        return RHO, dRHOdS, dRHOdT
    else:
        return RHO


@vectorize(
    ['float64(float64, float64, float64)', 'float32(float32, float32, float32)'],
    nopython=True,
)
def _compute_eos(salt, temp, pressure):
    # MWJF EOS coefficients
    # *** these constants will be used to construct the numerator
    mwjfnp0s0t0 = 9.99843699e2
    mwjfnp0s0t1 = 7.35212840
    mwjfnp0s0t2 = -5.45928211e-2
    mwjfnp0s0t3 = 3.98476704e-4
    mwjfnp0s1t0 = 2.96938239
    mwjfnp0s1t1 = -7.23268813e-3
    mwjfnp0s2t0 = 2.12382341e-3
    mwjfnp1s0t0 = 1.04004591e-2
    mwjfnp1s0t2 = 1.03970529e-7
    mwjfnp1s1t0 = 5.18761880e-6
    mwjfnp2s0t0 = -3.24041825e-8
    mwjfnp2s0t2 = -1.23869360e-11

    # *** these constants will be used to construct the denominator
    mwjfdp0s0t0 = 1.0
    mwjfdp0s0t1 = 7.28606739e-3
    mwjfdp0s0t2 = -4.60835542e-5
    mwjfdp0s0t3 = 3.68390573e-7
    mwjfdp0s0t4 = 1.80809186e-10
    mwjfdp0s1t0 = 2.14691708e-3
    mwjfdp0s1t1 = -9.27062484e-6
    mwjfdp0s1t3 = -1.78343643e-10
    mwjfdp0sqt0 = 4.76534122e-6
    mwjfdp0sqt2 = 1.63410736e-9
    mwjfdp1s0t0 = 5.30848875e-6
    mwjfdp2s0t3 = -3.03175128e-16
    mwjfdp3s0t1 = -1.27934137e-17

    salt2 = salt ** 0.5

    # compute density
    # *** first calculate numerator of MWJF density [P_1(S,T,p)]
    mwjfnums0t0 = mwjfnp0s0t0 + pressure * (mwjfnp1s0t0 + pressure * mwjfnp2s0t0)
    mwjfnums0t1 = mwjfnp0s0t1
    mwjfnums0t2 = mwjfnp0s0t2 + pressure * (mwjfnp1s0t2 + pressure * mwjfnp2s0t2)
    mwjfnums0t3 = mwjfnp0s0t3
    mwjfnums1t0 = mwjfnp0s1t0 + pressure * mwjfnp1s1t0
    mwjfnums1t1 = mwjfnp0s1t1
    mwjfnums2t0 = mwjfnp0s2t0

    WORK1 = (
        mwjfnums0t0
        + temp * (mwjfnums0t1 + temp * (mwjfnums0t2 + mwjfnums0t3 * temp))
        + salt * (mwjfnums1t0 + mwjfnums1t1 * temp + mwjfnums2t0 * salt)
    )

    # *** now calculate denominator of MWJF density [P_2(S,T,p)]
    mwjfdens0t0 = mwjfdp0s0t0 + pressure * mwjfdp1s0t0
    mwjfdens0t1 = mwjfdp0s0t1 + (pressure ** 3) * mwjfdp3s0t1
    mwjfdens0t2 = mwjfdp0s0t2
    mwjfdens0t3 = mwjfdp0s0t3 + (pressure ** 2) * mwjfdp2s0t3
    mwjfdens0t4 = mwjfdp0s0t4
    mwjfdens1t0 = mwjfdp0s1t0
    mwjfdens1t1 = mwjfdp0s1t1
    mwjfdens1t3 = mwjfdp0s1t3
    mwjfdensqt0 = mwjfdp0sqt0
    mwjfdensqt2 = mwjfdp0sqt2

    WORK2 = (
        mwjfdens0t0
        + temp * (mwjfdens0t1 + temp * (mwjfdens0t2 + temp * (mwjfdens0t3 + mwjfdens0t4 * temp)))
        + salt
        * (
            mwjfdens1t0
            + temp * (mwjfdens1t1 + temp * temp * mwjfdens1t3)
            + salt2 * (mwjfdensqt0 + temp * temp * mwjfdensqt2)
        )
    )

    DENOMK = 1.0 / WORK2

    return WORK1 * DENOMK


# @jit(nopython=True)
def _compute_eos_coeffs(salt, temp, pressure):
    # MWJF EOS coefficients
    # *** these constants will be used to construct the numerator
    mwjfnp0s0t0 = 9.99843699e2
    mwjfnp0s0t1 = 7.35212840
    mwjfnp0s0t2 = -5.45928211e-2
    mwjfnp0s0t3 = 3.98476704e-4
    mwjfnp0s1t0 = 2.96938239
    mwjfnp0s1t1 = -7.23268813e-3
    mwjfnp0s2t0 = 2.12382341e-3
    mwjfnp1s0t0 = 1.04004591e-2
    mwjfnp1s0t2 = 1.03970529e-7
    mwjfnp1s1t0 = 5.18761880e-6
    mwjfnp2s0t0 = -3.24041825e-8
    mwjfnp2s0t2 = -1.23869360e-11

    # *** these constants will be used to construct the denominator
    mwjfdp0s0t0 = 1.0
    mwjfdp0s0t1 = 7.28606739e-3
    mwjfdp0s0t2 = -4.60835542e-5
    mwjfdp0s0t3 = 3.68390573e-7
    mwjfdp0s0t4 = 1.80809186e-10
    mwjfdp0s1t0 = 2.14691708e-3
    mwjfdp0s1t1 = -9.27062484e-6
    mwjfdp0s1t3 = -1.78343643e-10
    mwjfdp0sqt0 = 4.76534122e-6
    mwjfdp0sqt2 = 1.63410736e-9
    mwjfdp1s0t0 = 5.30848875e-6
    mwjfdp2s0t3 = -3.03175128e-16
    mwjfdp3s0t1 = -1.27934137e-17

    salt2 = salt ** 0.5

    # compute density
    # *** first calculate numerator of MWJF density [P_1(S,T,p)]
    mwjfnums0t0 = mwjfnp0s0t0 + pressure * (mwjfnp1s0t0 + pressure * mwjfnp2s0t0)
    mwjfnums0t1 = mwjfnp0s0t1
    mwjfnums0t2 = mwjfnp0s0t2 + pressure * (mwjfnp1s0t2 + pressure * mwjfnp2s0t2)
    mwjfnums0t3 = mwjfnp0s0t3
    mwjfnums1t0 = mwjfnp0s1t0 + pressure * mwjfnp1s1t0
    mwjfnums1t1 = mwjfnp0s1t1
    mwjfnums2t0 = mwjfnp0s2t0

    WORK1 = (
        mwjfnums0t0
        + temp * (mwjfnums0t1 + temp * (mwjfnums0t2 + mwjfnums0t3 * temp))
        + salt * (mwjfnums1t0 + mwjfnums1t1 * temp + mwjfnums2t0 * salt)
    )

    # *** now calculate denominator of MWJF density [P_2(S,T,p)]
    mwjfdens0t0 = mwjfdp0s0t0 + pressure * mwjfdp1s0t0
    mwjfdens0t1 = mwjfdp0s0t1 + (pressure ** 3) * mwjfdp3s0t1
    mwjfdens0t2 = mwjfdp0s0t2
    mwjfdens0t3 = mwjfdp0s0t3 + (pressure ** 2) * mwjfdp2s0t3
    mwjfdens0t4 = mwjfdp0s0t4
    mwjfdens1t0 = mwjfdp0s1t0
    mwjfdens1t1 = mwjfdp0s1t1
    mwjfdens1t3 = mwjfdp0s1t3
    mwjfdensqt0 = mwjfdp0sqt0
    mwjfdensqt2 = mwjfdp0sqt2

    WORK2 = (
        mwjfdens0t0
        + temp * (mwjfdens0t1 + temp * (mwjfdens0t2 + temp * (mwjfdens0t3 + mwjfdens0t4 * temp)))
        + salt
        * (
            mwjfdens1t0
            + temp * (mwjfdens1t1 + temp * temp * mwjfdens1t3)
            + salt2 * (mwjfdensqt0 + temp * temp * mwjfdensqt2)
        )
    )

    DENOMK = 1.0 / WORK2

    RHOFULL = WORK1 * DENOMK

    # dRHOdT
    WORK3 = mwjfnums0t1 + temp * (2.0 * mwjfnums0t2 + 3.0 * mwjfnums0t3 * temp) + mwjfnums1t1 * salt

    WORK4 = (
        mwjfdens0t1
        + salt * mwjfdens1t1
        + temp
        * (
            2.0 * (mwjfdens0t2 + salt * salt2 * mwjfdensqt2)
            + temp * (3.0 * (mwjfdens0t3 + salt * mwjfdens1t3) + temp * 4.0 * mwjfdens0t4)
        )
    )

    DRHODT = (WORK3 - WORK1 * DENOMK * WORK4) * DENOMK

    # dRHOdS
    WORK3 = mwjfnums1t0 + mwjfnums1t1 * temp + 2.0 * mwjfnums2t0 * salt

    WORK4 = (
        mwjfdens1t0
        + temp * (mwjfdens1t1 + temp * temp * mwjfdens1t3)
        + 1.5 * salt2 * (mwjfdensqt0 + temp * temp * mwjfdensqt2)
    )

    DRHODS = (WORK3 - WORK1 * DENOMK * WORK4) * DENOMK * 1000.0

    return RHOFULL, DRHODS, DRHODT



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

import sys
from os.path import exists

from setuptools import setup

if exists('README.rst'):
    with open('README.rst') as f:
        long_description = f.read()
else:
    long_description = ''

with open('requirements.txt') as f:
    install_requires = f.read().strip().split('\n')

test_requirements = ['pytest']


CLASSIFIERS = [
    'Development Status :: 4 - Beta',
    'License :: OSI Approved :: Apache Software License',
    'Operating System :: OS Independent',
    'Intended Audience :: Science/Research',
    'Programming Language :: Python',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Topic :: Scientific/Engineering',
]

setup(
    maintainer='Matthew Long',
    maintainer_email='mclong@ucar.edu',
    description='POP2-CESM tools',
    install_requires=install_requires,
    license='Apache License 2.0',
    long_description=long_description,
    classifiers=CLASSIFIERS,
    keywords='ocean modeling cesm',
    name='pop-tools',
    packages=['pop_tools'],
    test_suite='tests',
    tests_require=test_requirements,
    include_package_data=True,
    url='https://github.com/NCAR/pop-tools',
    use_scm_version={'version_scheme': 'post-release', 'local_scheme': 'dirty-tag'},
    setup_requires=['setuptools_scm', 'setuptools>=30.3.0'],
    zip_safe=False,
)



"""Top-level module for pop_tools"""


from pkg_resources import DistributionNotFound, get_distribution

from .datasets import DATASETS
from .eos import compute_pressure, eos
from .fill import lateral_fill, lateral_fill_np_array
from .grid import get_grid, grid_defs
from .region_masks import list_region_masks, region_mask_3d
from .xgcm_util import to_xgcm_grid_dataset

try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    # package is not installed
    pass
